package org.mvcdi.component;

import java.lang.reflect.Method;
import java.util.Set;

import javax.enterprise.context.spi.CreationalContext;
import javax.enterprise.inject.spi.Bean;
import javax.enterprise.inject.spi.BeanManager;
import javax.inject.Inject;

import org.mvcdi.api.ControllerMethodMatcher;
import org.mvcdi.metadata.ControllerInfo;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class that is called from the servlet to handle requests.
 * 
 * @author Andy Gibson
 * 
 */
public class MvcHandler {

	private static final Logger logger = LoggerFactory
			.getLogger(MvcHandler.class);

	@Inject
	private ControllerInfo controllerInfo;

	@Inject
	private ControllerMethodMatcher matcher;

	@Inject
	private BeanManager beanManager;

	public void initialize() {
		controllerInfo.initialize();
	}

	public void handleRequest(RequestContext context) {
		context.setMethod(matcher.findMatching(controllerInfo, context));
		if (context.getMethod() != null) {
			Object controller = locateController(context.getMethod().getControllerClass());
			context.setController(controller);
			// start executing the controller method
			executeControllerMethod(context);
			handleResponse(context);
		} else {
			throw new RuntimeException("Unable to find method for "
					+ context.getRequestMethod() + " request with url "
					+ context.getPath());
		}
	}

	private void executeControllerMethod(RequestContext context) {
		Method javaMethod = context.getMethod().getMethod();
		Object outcome = null;		
		try {
			outcome = javaMethod.invoke(context.getController());
		} catch (Exception e) {
			e.printStackTrace();
		}
		context.setOutcome(outcome);
	}

	private void handleResponse(RequestContext context) {
		if (context.getOutcome() instanceof String) {
			String outcome = (String) context.getOutcome();
			String view = "/" + outcome + ".jsp";
			logger.debug("resolved outcome {} to view {}", outcome, view);
			context.forwardTo(view);
		}
	}

	private Object locateController(Class<?> controllerClass) {
		// locate the controller class from CDI
		Set<Bean<?>> beans = beanManager.getBeans(controllerClass,
				ControllerInfo.CONTROLLER_LITERAL);
		if (beans.isEmpty()) {
			throw new RuntimeException(
					"Unable to locate instance of the controller of type "
							+ controllerClass);
		}

		Bean<?> bean = beanManager.resolve(beans);
		CreationalContext<?> context = beanManager
				.createCreationalContext(bean);

		Object result = beanManager.getReference(bean, bean.getBeanClass(),
				context);
		return result;
	}

}
