package org.mvcdi.metadata;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;

import javax.annotation.PostConstruct;
import javax.enterprise.context.ApplicationScoped;
import javax.enterprise.inject.spi.Bean;
import javax.enterprise.inject.spi.BeanManager;
import javax.enterprise.util.AnnotationLiteral;
import javax.inject.Inject;

import org.mvcdi.annotations.Controller;
import org.mvcdi.annotations.RequestMapping;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Preconditions;

@ApplicationScoped
public class ControllerInfo {

	private final List<ControllerMethod> controllerMethods = new ArrayList<ControllerMethod>();

	private static final String[] DEFAULT_MAPPING_PATHS = new String[] { "" };

	public static final AnnotationLiteral<Controller> CONTROLLER_LITERAL = new AnnotationLiteral<Controller>() {
		private static final long serialVersionUID = -3226395594698453241L;
	};

	@Inject
	private BeanManager beanManager;
	
	@PostConstruct
	public void initialize() {
		logger.debug("Initializing controller info");
		Set<Bean<?>> controllers = beanManager.getBeans(Object.class,
				CONTROLLER_LITERAL);
		for (Bean<?> bean : controllers) {
			add(bean.getBeanClass());
		}
		sortControllerMethods();
	}

	private final static Logger logger = LoggerFactory
			.getLogger(ControllerInfo.class);

	public List<ControllerMethod> getControllerMethods() {
		return controllerMethods;
	}

	private void add(Class<?> clazz) {
		logger.debug("Adding class {}", clazz);
		Preconditions.checkNotNull(clazz);
		if (clazz.isAnnotationPresent(Controller.class)) {
			logger.debug("Found controller on class {}", clazz);
			RequestMapping rm = clazz.getAnnotation(RequestMapping.class);

			Method[] methods = clazz.getMethods();
			String[] controllerPaths = null;
			if (rm != null) {
				controllerPaths = rm.value();
			}
			// if no paths are specified, then default to one blank path
			if (controllerPaths == null || controllerPaths.length == 0) {
				controllerPaths = DEFAULT_MAPPING_PATHS;
			}
			// add methods for each controller level paths
			for (String prefix : controllerPaths) {
				for (Method m : methods) {
					addMethod(prefix, m);
				}
			}
		}
	}

	private void addMethod(String prefix, Method javaMethod) {
		Preconditions.checkNotNull(javaMethod);
		RequestMapping mapping = javaMethod.getAnnotation(RequestMapping.class);
		if (mapping != null) {
			logger.debug("Found request mapping on method {}", javaMethod
					.getName());
			String[] paths = mapping.value();

			// if these are blank, fill with defaults
			if (paths.length == 0) {
				paths = DEFAULT_MAPPING_PATHS;
			}

			for (String path : paths) {
				controllerMethods.add(new ControllerMethod(javaMethod, prefix,
						path, mapping.method()));
			}

		}
		// check for other annotations in the future
	}

	public void listMethods() {
		System.out.println("Method count = " + controllerMethods.size());
		for (ControllerMethod method : controllerMethods) {
			System.out.println(method);
		}
	}

	private void sortControllerMethods() {		
		Comparator<ControllerMethod> sorter = new Comparator<ControllerMethod>() {
			public int compare(ControllerMethod cm1, ControllerMethod cm2) {
				return cm2.getPathLength() - cm1.getPathLength();
			}
		};
		Collections.sort(controllerMethods, sorter);
	}
}
